#pragma once

/*
MIT License
Copyright (c) 2024 René Amthor (tobid7)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
 */

#include <pd/common/common.hpp>

using UI7Color = PD::u32;

enum UI7Color_ {
  UI7Color_Background,
  UI7Color_Button,
  UI7Color_ButtonDead,
  UI7Color_ButtonActive,
  UI7Color_ButtonDisabled,
  UI7Color_Text,
  UI7Color_TextDead,
  UI7Color_Header,
  UI7Color_Selector,
  UI7Color_Checkmark,
  UI7Color_FrameBackground,
  UI7Color_FragmeBackgroundHovered,
  UI7Color_Progressbar,
  UI7Color_ListEven,
  UI7Color_ListOdd,
};

namespace PD {
namespace UI7 {
/// @brief Theme Class
class Theme {
 public:
 public:
  Theme() { Default(*this); }
  ~Theme() {}

  /// @brief Simple static Loader for the Default Theme
  /// @param theme Theme Reference
  static void Default(Theme& theme);

  /// @brief Revert the last Color Change
  Theme& Pop() {
    theme[changes[changes.size() - 1].first] =
        changes[changes.size() - 1].second;
    changes.pop_back();
    return *this;
  }

  /// @brief Revert the last color Change done for a specific color
  /// @param c Color to revert change from
  Theme& Pop(UI7Color c) {
    for (size_t i = changes.size() - 1; i > 0; i--) {
      if (changes[i].first == c) {
        theme[c] = changes[i].second;
        changes.erase(changes.begin() + i);
        break;
      }
    }
    return *this;
  }

  /// @brief Change a Color
  /// @param tc Color Identifier
  /// @param color Color to change to
  Theme& Change(UI7Color tc, u32 color) {
    if (theme.find(tc) == theme.end()) {
      return *this;
    }
    changes.push_back(std::make_pair(tc, theme[tc]));
    theme[tc] = color;
    return *this;
  }

  /// @brief Get the Color of a Color ReferenceID
  /// @param c ReferenceID
  u32 Get(UI7Color c) const {
    auto e = theme.find(c);
    if (e == theme.end()) {
      return 0x00000000;
    }
    return e->second;
  }

  /// @brief Operator wrapper for get
  /// @param c Color ReferenceID
  u32 operator[](UI7Color c) const { return Get(c); }

  /// @brief Change but just sets [can implement completly new ids]
  /// @param tc Color ID (Can be self creeated ones as well)
  /// @param clr Color it should be set to
  void Set(UI7Color tc, u32 clr) { theme[tc] = clr; }

 private:
  std::unordered_map<u32, u32> theme;
  std::vector<std::pair<UI7Color, u32>> changes;
};
}  // namespace UI7
/// Using UI7Color as a Class to be able to
/// define it as struct as well as using it as enum
class UI7Color {
 public:
  UI7Color() {
    /// No Color
  }
  UI7Color(unsigned int c) { color = c; }
  UI7Color(UI7Color_ c) {}
  ~UI7Color() {}

  operator u32() const { return color; }

 private:
  u32 color;
};
}  // namespace PD